package application;

import debug.*;
import timer.*;
import link.*;
import network.*;
import reliability.*;

import java.io.*;
import java.util.*;

/***************************************************************** 
  A class to act as an application to drive the sending end of a 
  bytestream conversation.  Matches up with a 
  BytestreamReceiverClient on the other end, and works through
  a ReliableSender instance to acheive communication.
  <p>
  You probably won't need to worry about this class, since it 
  will be set up by the config file parser.
 *****************************************************************/
public class BytestreamSenderClient extends TimerClient 
                                    implements Runnable
{
   private ReliableSender sender;
   private Vector uptime;
   private Address addr;
   private Address receiverAddr;

   /************************************************************
     Creates a new sender client and starts the run thread for 
     it.  Clients should not be made to share sender instances,
     because it probably won't work.

     @param sendImplementation the ReliableSender instance
            (which needs to be connected to a network and have
            an address set) through which to transmit

     @param upPeriods a vector of <tt>TimeInterval</tt>s
            indicating which time periods data should be sent
            during.  If <tt>null</tt> then this sender will
            transmit continuously forever

     @param address the address of this host.  Used to name
            the sending thread

     @param receiverAddress the address of the receiver, used
            to open the connection
    ************************************************************/
   public BytestreamSenderClient (ReliableSender sendImplementation, Vector upPeriods,
                                  Address address, Address receiverAddress)
   {
      sender = sendImplementation;
      uptime = upPeriods;
      addr = address;
      receiverAddr = receiverAddress;

      Debug.println ('a', "Sender application at " + address + 
                          " with interval list:");
      Enumeration e = uptime.elements ();
      while (e.hasMoreElements ())
         {
         TimeInterval t = (TimeInterval) e.nextElement ();
         Debug.println ('a', "  [" + t.start + ", " + t.end + "]");
         }

      Thread sendThread = new Thread (this);
      Timer.global.RegisterThread (sendThread);
      sendThread.start ();
   }

   /************************************************************
     The sending thread.  Clients should not call this: the
     thread is started automatically by the constructor
    ************************************************************/
   public void run ()
   {
      Thread.currentThread ().setName ("Sender application at " + addr);
      synchronized (Timer.global)
         {
         sender.Open (receiverAddr);

         byte[] data = new byte[50];
         String s = "";
         int count = 0;
         while (uptime == null || uptime.size () > 0)
            {
            boolean transmit = true;
            if (uptime != null)
               {
               Debug.print ('a', "Interval list at " + addr + 
                                   ": ");
               Enumeration e = uptime.elements ();
               while (e.hasMoreElements ())
                  {
                  TimeInterval t = (TimeInterval) e.nextElement ();
                  Debug.print ('a', "  [" + t.start + ", " + t.end + "]");
                  }
               Debug.println ('a', "");
              
               TimeInterval current = (TimeInterval) uptime.elementAt (0);
               if (current.start > Timer.global.CurrentSimTime ())
                  {
                  Debug.println ('a', "" + Timer.global.CurrentSimTime () + 
                                      ": " + addr + 
                     ": send waiting for start at " + current.start);
                  
                  Timer.global.WaitUntil (this, current.start, Timer.US);
                  }
               else if (current.end < Timer.global.CurrentSimTime ())
                  {
                  Debug.println ('a', "" + Timer.global.CurrentSimTime () + 
                                      ": " + addr + 
                     ": cleared end at " + current.end + 
                     ", send advancing to next period");

                  uptime.removeElementAt (0);
                  transmit = false;
                  }
               else 
                  Debug.println ('a', "" + Timer.global.CurrentSimTime () +
                                      ": " + addr + 
                             ": continuing in interval [" +
                             current.start + ", " + current.end + "]");
               }
            else
               Debug.println ('a', "" + Timer.global.CurrentSimTime () +
                                   ": " + addr + ": uptime is null");

            if (transmit)
               {
               while (s.length () < 50)
                  {
                  s = s + " " + count;
                  ++count;
                  }

               s.getBytes (0, 50, data, 0);
               s = s.substring (50);

               sender.Send (data);
               }
            }

         Timer.global.UnregisterThread (Thread.currentThread ());
         Debug.println ('a', "Send application at " + addr + " shuting down");
         }
   }
}

