package network;

import java.util.*;
import java.io.*;
import timer.*;
import debug.*;
import link.*;
   
/***************************************************************** 
   The implementation of a router.  Routers connect to multiple
   links (either point-to-point or broadcast), and forward
   packets between them.
   <p>
   The basic implementation provided maintains a FIFO queue of
   packets to be dealt with, and forwards packets to all 
   network links other than the one they arrived on.  If the
   queue overflows then packets are dropped.  This implementation
   also names each router so that debug messages can make sense.
   <p>
   You will need to add code to enable routers, by cooperating
   with each other and with the RoutedNetworkEdges at the
   edges of the network, to discover the network topology and
   route packets more efficiently (at the very least your routers 
   should forward packets out only one link, and maintain the
   property that they make it to their destinations.  
   <p>
   You will most likely need to modify <tt>run</tt> to handle 
   forwarding, and either <tt>PacketArrived</tt> or <tt>run</tt> 
   to deal with packets for your topology discovery.  If you need
   to add fields to packets (you will) you should modify the
   class <tt>NetworkPacket</tt>.

   @see Node
   @see RoutedNetworkEdge
   @see NetworkPacket
   @see Link
 *****************************************************************/
public class Router extends TimerClient implements Node, Runnable
{
   /************************************************************
      Create a new node.
 
      @param name the name of this router
      @param queueLen the number of packets to queue, in 
             packets
    ************************************************************/
   public Router (String name, int queueLen)
   {
      this.name = name;
      links = new Vector ();
      queue = new Vector ();
      maxQueueSize = queueLen;

      // Start a thread to forward packets.
      forwardThread = new Thread (this);
      Timer.global.RegisterThread (forwardThread);
      forwardThread.start ();
   }

   /************************************************************
      The callback for new packets.  Appends the new packet to
      the queue if there is space, otherwise the packet is 
      dropped.

      @see Node#PacketArrived
    ************************************************************/
   public void PacketArrived (LinkPacket p, Link l)
   {
      // Assume that the packet is of the network flavor
      NetworkPacket packet = (NetworkPacket) p;

      if (queue.size () < maxQueueSize)
         {
         queue.addElement (new RouterPacketRecord (packet, l));
         Debug.println ('R', "" + Timer.global.CurrentSimTime () +
                        ": " + this + ": Added packet " + packet + 
                        " to queue, queue length now " +
                        queue.size () + " (of " + maxQueueSize + ")");
         Notify ();
         }
      else
         Debug.println ('R', "" + Timer.global.CurrentSimTime () +
                        ": " + this + ": Dropped packet " + packet);
   }

   /************************************************************
      The forwarding thread.  Continually gets packets from 
      the front of the queue and fowards them.
      <p>
      The provided algorithm forwards packets to all other 
      links, and also implements a time-to-live field, 
      which limits the number of hops.  This field is added
      in <tt>NetworkPacket</tt>, and a constructor is used
      there to set the default ttl to 10 hops.

      @see NetworkPacket
    ************************************************************/
   public void run ()
   {
      synchronized (Timer.global)
         {
         for (;;)
            {
            Enumeration e;
            RouterPacketRecord rec;

            // Wait for more packets to arrive if there are none.
            while (queue.size () == 0)
               Wait ();

            // Forward the first packet
            rec = (RouterPacketRecord) queue.elementAt (0);
            queue.removeElementAt (0);

            // See if ttl has expired, and update it
            if (rec.packet.timeToLive <= 0)
               continue;          // go on around the loop for another packet
            --rec.packet.timeToLive;
               
            // Get the list of connected links
            e = ConnectedLinks ();

            // Retransmit the packet
            while (e.hasMoreElements ())
               {
               Link link = (Link) e.nextElement ();
               if (link != rec.arrivedOn)
                  link.TransmitPacket (rec.packet, this);
               }
            }
         }
   }

   private Vector queue;       // holds RouterPacketRecords
   private int maxQueueSize;
   private Thread forwardThread;

   //---------------------------------------------------------
   // Stuff below you probably won't need to modify, but you
   // will probably want to use ConnectedLinks at some point.

   /************************************************************
      Get an enumeration of the currently connected links.
      @return an enumeration of the links
      @see Enumeration
    ************************************************************/
   protected Enumeration ConnectedLinks ()
   {
      return links.elements ();
   }

   /************************************************************
      Connect a link
      @see Node#AddLink
    ************************************************************/
   public void AddLink (Link l)
   {
      if (!links.contains (l))
         links.addElement (l);
   }

   /************************************************************
      Disconnect a link
      @see Node#RemoveLink
    ************************************************************/
   public void RemoveLink (Link l)
   {
      links.removeElement (l);
   }

   /************************************************************
      String representation is name plus type:
    ************************************************************/
   public String toString ()
   {
      return "Router:" + name;
   }

   private String name;
   private Vector links;
}

/***************************************************************** 
   A structure to keep information about a packet that is 
   being held at a router
 *****************************************************************/
class RouterPacketRecord
{
   public RouterPacketRecord (NetworkPacket p, Link l)
   {
      arrivedOn = l;
      packet = p;
   }

   public NetworkPacket packet;
   public Link arrivedOn;
}

