package reliability;

import debug.*;
import timer.*;
import java.util.*;
import java.io.*;

/***************************************************************** 
   A specialized bounded-buffer for receiving.  This 
   buffer has the semantics that removes will wait
   for data to become available but appends will 
   always return immediately
 *****************************************************************/
public class ReceiveBuffer extends TimerClient
{
   /************************************************************
      Set up the buffer with the given size.  Needs
      the global timer to ensure that waiting threads
      are properly woken

      @param size the size of the buffer in bytes
    ************************************************************/
   public ReceiveBuffer (int size)
   {
      buffer = new byte[size];
      numBytes = 0;
   }

   /************************************************************
      Append data to the end of the buffer/queue.
      If there is no room, then fail and return false

      @param data the data to append
    ************************************************************/
   public boolean AppendBytes (byte[] data)
   {
      boolean retval = true;

      if (numBytes + data.length < buffer.length)
         {
         for (int i=0; i<data.length; ++i)
            buffer[numBytes + i] = data[i];

         numBytes += data.length;   
         }
      else
         retval = false;

      NotifyAll ();

      return retval;
   }

   /************************************************************
      Remove maxBytes from the buffer.  If less
      or no data is available, then block until data
      becomes available.  

      @param maxBytes the size to return
      @return the data removed
    ************************************************************/
   public byte[] RemoveBytes (int maxBytes)
   {
      while (maxBytes > numBytes)
         Wait ();

      byte[] retval = new byte[maxBytes];
      for (int i=0; i<maxBytes; ++i)
         retval[i] = buffer[i];

      for (int i=maxBytes; i<numBytes; ++i)
         buffer[i-maxBytes] = buffer[i];
      numBytes -= maxBytes;

      return retval;
   } 

   private byte[] buffer;
   private int numBytes;
}

