package reliability;
import debug.*;
import timer.*;
import network.*;

import java.io.*;
import java.util.*;

/***************************************************************** 
   This class implements a stop-and-wait reliability algorithm.
   Connection setup is ignored.  This will cause problems with
   connection patterns involving multiple senders.

   @see ReliableReceiver
 *****************************************************************/
public class SAWReceiver implements ReliableReceiver, PacketReceiver
{
   /************************************************************
      Set up a receiver.  

      @param network the network over which to communicate
      @param address the address of this host
    ************************************************************/
   public SAWReceiver (Network network, Address address)
   {
      this.address = address;

      buffer = new ReceiveBuffer (50 * network.MTU ());
      this.network = network;
      network.AddReceiver (this, address);

      currentSequenceNum = 1;       // stop-and-wait setup
   }

   private Network network;
   private Address address;
   private ReceiveBuffer buffer;

   private int currentSequenceNum;   // for stop-and-wait

   /************************************************************
      Receive a block of data from the connection.  Data blocks 
      of any size may be received, and the operation will block
      until that much data is available.  

      @param size the number of bytes requested
      @return the data received, up to the number of bytes
              requested
    ************************************************************/
   public byte[] Receive (int numBytes)
   {
      return buffer.RemoveBytes (numBytes);
   }

   /************************************************************
      Processing that happens when we receive a packet
      <p>
      For stop-and-wait: ack iff packet is old or can be stored,
      advance currentSequenceNum if packet is new and stored
    ************************************************************/
   public void PacketArrived (NetworkPacket pk)
   {
      ReliablePacket p = (ReliablePacket) pk;

      if (p.sequenceNum > currentSequenceNum)
         return;

      if (p.sequenceNum == currentSequenceNum)
         {
         if (!buffer.AppendBytes (p.data))
            {
            System.out.println ("Dropping packet: no space");
            return;   // no space, cannot ack
            }

         ++currentSequenceNum;
         }

      // send ack for this packet
      ReliablePacket ack = new ReliablePacket ();
      ack.sourceAddr = address;
      ack.destinationAddr = p.sourceAddr;

      ack.sequenceNum = p.sequenceNum;
      ack.ack = true;
      
      ack.data = new byte[0];

      network.TransmitPacket (ack);      
   }
}
 

