package reliability;

import debug.*;
import timer.*;
import network.*;

import java.io.*;
import java.util.*;

/***************************************************************** 
   Implementation of a stop-and-wait sender
   <p>
   Provided implementation shows stop-and-wait protocol,
   plus all the general setup and use of the SendBuffer
   <p>
   This implementation uses both the SendBuffer and
   NetworkInterface classes.  You are not required to use
   these, but they simplify interaction with the timer and
   the underlying network.  If you want to modify or replace
   them, be sure to understand them first.
 *****************************************************************/
public class SAWSender extends TimerClient
                implements ReliableSender, PacketReceiver
{
   /************************************************************
      Create a new stop-and-wait sender.
 
      @param network the network to communicate on
      @param address the address of this host
    ************************************************************/
   public SAWSender (Network network, Address address)
   {
      this.address = address;
      destinationAddress = null;

      this.network = network;
      network.AddReceiver (this, address);

      buffer = new SendBuffer (50 * network.MTU ());

      // The following is for stop-and-wait, feel free to remove
      currentPacket = null;
      currentSequenceNum = 0;
      currentTimeout = null;
   }

   private Network network;
   private Address address;
   private SendBuffer buffer;
   private Address destinationAddress;

   /************************************************************
      Open a connection to the ReliableReceiver represented by
      the address.  This call blocks until the connection is
      established.
      <p>
      For stop-and-wait, true connection setup is not implemented,
      and this jsut sets the destination
 
      @param destination the address of the receiver
    ************************************************************/
   public void Open (Address destination)
   {
      destinationAddress = destination;
   }

   /************************************************************
      Send a block of data on the connection.  Any amount of
      data may be sent.  This call may block if there is 
      insufficient buffer space.  Clients may make any number
      of <tt>Send</tt> calls per open connection.

      @param data the data (any amount) to send on the connection
      @throws ConnectionNotOpenException if the connection is
          not open
    ************************************************************/
   public void Send (byte[] data) throws ConnectionNotOpenException
   {
      if (destinationAddress == null)
         throw new ConnectionNotOpenException ();

      int previous = buffer.AppendBytes (data);

      if (previous == 0)      // Not currently sending
         SendNextPacket ();   // so start again
   }

   /************************************************************
      Close the current connection.  Blocks until all data
      is read by the other application and the close
      operation completes.

      @throws ConnectionNotOpenException if the connection is
          not open
    ************************************************************/
   public void Close () throws ConnectionNotOpenException
   {
      if (destinationAddress == null)
         throw new ConnectionNotOpenException ();

      destinationAddress = null;
   }

   /************************************************************
      Processing that happens when we receive a packet
      <p>
      For stop-and-wait: send the next packet
    ************************************************************/
   public void PacketArrived (NetworkPacket pk)
   {
      ReliablePacket p = (ReliablePacket) pk;
      
      if (p.sequenceNum == currentSequenceNum)
         SendNextPacket ();
   }

   /************************************************************
      Processing that happens when we get a timeout
      <p>
      For stop-and-wait: resend the current packet
    ************************************************************/
   public void Timeout (Object info)
   {
      SendCurrentPacket ();
   }

   private ReliablePacket currentPacket;
   private int currentSequenceNum;
   private TimeoutID currentTimeout;

   /************************************************************
    ************************************************************/
   private void SendNextPacket ()
   {
      byte[] data = buffer.RemoveBytes (network.MTU ());

      if (data.length > 0)
         {
         currentPacket = new ReliablePacket ();

         currentPacket.sourceAddr = address;
         currentPacket.destinationAddr = destinationAddress;

         ++currentSequenceNum;
         currentPacket.sequenceNum = currentSequenceNum;
         currentPacket.ack = false;

         currentPacket.data = data;

         if (currentTimeout != null)
            Timer.global.CancelTimeout (currentTimeout);

         network.TransmitPacket (currentPacket);
         currentTimeout = Timer.global.SetTimeout (this, 1000, Timer.US, null);
         }
   }

   /************************************************************
    ************************************************************/
   private void SendCurrentPacket ()
   {
      if (currentPacket != null)
         {
         network.TransmitPacket (currentPacket);
         }

      currentTimeout = Timer.global.SetTimeout (this, 1000, Timer.US, null);
   }
  
}


