package reliability;

import debug.*;
import timer.*;
import network.*;

import java.io.*;
import java.util.*;

/***************************************************************** 
   A specialized bounded-buffer for sending.  This 
   buffer has the semantics that appends will wait for
   space but removes will always return immediately.
 *****************************************************************/
public class SendBuffer  extends TimerClient
{
   /************************************************************
      Create a new SendBuffer

      @param size size of buffer in bytes
    ************************************************************/
   public SendBuffer (int size)
   {
      buffer = new byte[size];
      numBytes = 0;
   }

   /************************************************************
      Append data to the end of the buffer/queue.
      If there is no room, then block until space
      becomes available.

      @return the old number of bytes in the buffer (before the
                  insertion)
    ************************************************************/
   public int AppendBytes (byte[] data)
   {
      while (numBytes + data.length > buffer.length)
         Wait ();

      for (int i=0; i<data.length; ++i)
         buffer[numBytes + i] = data[i];

      int retval = numBytes;
      numBytes += data.length;         

      return retval;
   }

   /************************************************************
      Remove up to maxBytes from the buffer.  If less
      or no data is available, then return what is
      available (may return a zero-length array)
    ************************************************************/
   public byte[] RemoveBytes (int maxBytes)
   {
      int toRemove = Math.min (maxBytes, numBytes);

      byte[] retval = new byte[toRemove];
      for (int i=0; i<toRemove; ++i)
         retval[i] = buffer[i];

      for (int i=toRemove; i<numBytes; ++i)
         buffer[i-toRemove] = buffer[i];
      numBytes -= toRemove;

      NotifyAll ();

      return retval;
   } 

   public int Size ()
   {
      return numBytes;
   }

   private byte[] buffer;
   private int numBytes;
}

