package reliability;
import debug.*;
import timer.*;
import network.*;

import java.io.*;
import java.util.*;

/***************************************************************** 
   This class implements a sliding-window reliability algorithm

   @see ReliableReceiver
 *****************************************************************/
public class SlidingWindowReceiver implements ReliableReceiver, PacketReceiver
{
   /************************************************************
      Set up a receiver.  

      @param network the network over which to communicate
      @param address the address of this host
    ************************************************************/
   public SlidingWindowReceiver (Network network, Address address)
   {
      this.address = address;

      buffer = new ReceiveBuffer (50 * network.MTU ());
      this.network = network;
      network.AddReceiver (this, address);

      window = null;
   }

   private Network network;
   private Address address;
   private ReceiveBuffer buffer;

   private ReceiveWindow window;

   /************************************************************
      Receive a block of data from the connection.  Data blocks 
      of any size may be received, and the operation will block
      until that much data is available.  

      @param size the number of bytes requested
      @return the data received, up to the number of bytes
              requested
    ************************************************************/
   public byte[] Receive (int numBytes)
   {
      return buffer.RemoveBytes (numBytes);
   }

   /************************************************************
      Processing that happens when we receive a packet
      <p>
      For stop-and-wait: ack iff packet is old or can be stored,
      advance currentSequenceNum if packet is new and stored
    ************************************************************/
   public void PacketArrived (NetworkPacket pk)
   {
      ReliablePacket p = (ReliablePacket) pk;

      if (p.syn)
         {
         if (window == null)
            {
            int isn = p.sequenceNum+1;
            window = new ReceiveWindow (SlidingWindowSender.WINDOW_SIZE + 5, isn);
            }

         ReliablePacket ack = new ReliablePacket ();
         ack.sourceAddr = address;
         ack.destinationAddr = p.sourceAddr;

         ack.sequenceNum = p.sequenceNum;
         ack.ack = true;
         ack.syn = true;

         ack.data = new byte[0];

         network.TransmitPacket (ack);

         return;
         }

      if (window == null)
         return;

      Debug.println ('r', "" + Timer.global.CurrentSimTime () +
                          ": Packet arrived at receiver: " + p.sequenceNum);

      boolean packetAccepted = true;
      int ackSequenceNum = -1;

      // If this is so far ahead that it exceeds the window, then drop
      if (p.sequenceNum > window.SequenceNumAt (window.Size () - 1))
         packetAccepted = false;

      // Otherwise, if this belongs in the window, then grab it
      else if (p.sequenceNum >= window.SequenceNumAt (0))
         window.GrabPacket (p);

      // Okay: shift all packets we can into the receive buffer
      //       this loop has the property that it shifts until
      //       it reaches a missing packet (normal case)
      //       or the receive buffer fills up (application too
      //       slow).  If the app is too slow, this will lead
      //       to us running out of window space, which is okay.
      while (window.PacketAt (0) != null)
         {
         if (!buffer.AppendBytes (window.PacketAt (0).data))
            break;
         window.Rotate ();
         }

      Debug.print ('r', "" + Timer.global.CurrentSimTime () + ": Receive window:");
      for (int i=0; i<window.Size (); ++i)
         if (window.PacketAt (i) == null)
            Debug.print ('r', " (" + window.SequenceNumAt (i) + ")");
         else
            Debug.print ('r', "  " + window.SequenceNumAt (i) + " ");
      Debug.println ('r', "");

      // If ack for all previous then find the first missing packet
      // Assume this is the first packet in the window.  The exception
      // is if the receive buffer filled, but this will be okay.
      if (SlidingWindowSender.ACK_FOR_ALL_PREVIOUS)
         ackSequenceNum = window.SequenceNumAt (0) - 1;

      // Otherwise ack this packet iff we accepted it
      else if (packetAccepted)
         ackSequenceNum = p.sequenceNum;

      if (!packetAccepted)
         return;             // drop w/o acking

      // send ack for this packet
      ReliablePacket ack = new ReliablePacket ();
      ack.sourceAddr = address;
      ack.destinationAddr = p.sourceAddr;

      ack.sequenceNum = ackSequenceNum;
      ack.ack = true;
      ack.syn = false;

      ack.data = new byte[0];

      network.TransmitPacket (ack);      
   }
}
 

class ReceiveWindow
{
   public ReceiveWindow (int N, int isn)
   {
      window = new ReceiveWindowData[N];
      for (int i=0; i<N; ++i)
         {
         window[i] = new ReceiveWindowData ();
         window[i].packet = null;
         window[i].sequenceNum = i+isn;
         }
      windowStart = 0;
   }

   public int Size ()
   {
      return window.length;
   }

   public ReliablePacket PacketAt (int offset)
   {
      return DataAtOffset (offset).packet;
   }

   public int SequenceNumAt (int offset)
   {
      return DataAtOffset (offset).sequenceNum;
   }

   public void GrabPacket (ReliablePacket p)
   {
      for (int i = 0; i < Size (); ++i)
         if (SequenceNumAt (i) == p.sequenceNum)
            DataAtOffset (i).packet = p;
   }

   public void Rotate ()
   {
      window[windowStart].packet = null;
      window[windowStart].sequenceNum += window.length;
      windowStart = (windowStart + 1) % window.length;
   }

   // Representation: circular queue: windowStart is the index
   // of item 0, items go up from there and wrap around.
   private ReceiveWindowData[] window;
   private int windowStart;

   private ReceiveWindowData DataAtOffset (int offset)
   {
      return window[(offset + windowStart) % window.length];
   }
}

class ReceiveWindowData
{
   public ReliablePacket packet;
   public int sequenceNum;
}
