//-----------------------------------------------------------------
// CSE 461 Project 1:                        Timer implementation
//
// implementation of the simulated time clock
// and other time-related functions.
//-----------------------------------------------------------------
package timer;

import java.io.*;
import java.util.*;
import debug.*;


/***************************************************************** 
   The simulated-time clock.  This is a large class, and
   the interface is divided into several sections:
   <ul>
   <li> <b>Clock interface</b>:    read the time
        <ul> 
        <li> <tt>public void CurrentSimTime ()</tt>
        </ul>

   <li> <b>Timeout interface</b>:  establish timeouts that will fire some
                                   time in the future
        <ul> 
        <li> <tt>public TimeoutID SetTimeout ()</tt>
        <li> <tt>public void CancelTimeout (TimeoutID)</tt>
        </ul>

   <li> <b>Wait interface</b>:     block the current thread until some
                                   time in the future.
        <ul>
        <li> <tt>public void WaitFor (TimerClient, long, TimeUnit)</tt>
        <li> <tt>public void WaitUntil (TimerClient, long, TimeUnit)</tt>
        </ul>
   </ul>
   
   The threading interaction between clients and the timer 
   is nontrivial.  The following rules of thumb should be 
   observed.
   <ul>
   <li> All client classes should <tt>extend</tt> the class
        <tt>TimerClient</tt>.  This identifies them as clients
        and gives them access to monitor routines that work
        correctly with the timer.

   <li> Clients which only want to set and receive timeouts should
        have little trouble.  They will need to override the 
        <tt>Timeout</tt> function with one of their own, and should
        generally avoid blocking or infinite processing inside
        this function.  If they do need to block, they should use
        either the modified routines in <tt>TimerClient</tt> or
        the <tt>WaitFor</tt> or <tt>WaitUntil</tt> functions in
        the timer.  They should avoid using the builtin Java
        blocking routines, as this typically leads to deadlock.

   <li> Clients which want to implement traditional multithreaded
        applications need to register and unregister all their
        threads with the timer.  Threads should remain registered
        throughout their entire lifetime, but <i>must</i> be 
        unregistered when they terminate.  Otherwise, these
        clients should observe the same rules as timeout clients.
   </ul>
 *****************************************************************/
public class Timer implements Runnable
{
   /************************************************************
    ************************************************************/

   /************************************************************
      The time unit seconds.  Time is maintained internally
      in microseconds, but these units allow more convenient
      specification of time periods in the public functions.
    ************************************************************/
   public static TimeUnit SEC = new TimeUnit (1000000);

   /************************************************************
      The time unit milliseconds
    ************************************************************/
   public static TimeUnit MS  = new TimeUnit (1000);

   /************************************************************
      The time unit microseconds
    ************************************************************/
   public static TimeUnit US  = new TimeUnit (1);


   /************************************************************
      The global timer.  Set up at program startup
    ************************************************************/
   public static Timer global = new Timer ();

   /************************************************************
      Sets stuff up and starts the event processing thread.  
      Clients are not allowed to create new timers.  They
      should use the global timer <tt>Timer.global</tt>

      @see Timer#global
    ************************************************************/
   Timer ()
   {
      simTime = 0;
      activeThreads = new Hashtable ();
      pendingTimeouts = new Vector ();

      timeoutThread = new Thread (this);
      timeoutThread.start ();

      (new TimerMonitor (this)).start ();
   }

   /************************************************************
      Stops the timer, not really necessary.
    ************************************************************/
   public void Shutdown ()
   {
      timeoutThread.stop ();
      timeoutThread = null;
      activeThreads = null;
      pendingTimeouts = null;
   }

   /************************************************************
      Return the current simulator time in microseconds
    ************************************************************/
   public long CurrentSimTime ()
   {
      return simTime;
   }




   /************************************************************
      SetTimeout:  set a timeout for the future

      @param client  the client to wake up when the time expires

      @param time    the number of time units from now that the
                     timeout will expire

      @param units   the units of time, one of
                     <ul>
                     <li>  <tt>Timer.SEC</tt>: seconds
                     <li>  <tt>Timer.MS</tt>: milliseconds
                     <li>  <tt>Timer.US</tt>: microseconds
                     </ul>

      @param info    the information about the timeout to pass
                     back to the client when the timeout expires

      @see TimerClient#Timeout
    ************************************************************/
   public synchronized TimeoutID SetTimeout (TimerClient client, 
                                             long time, 
                                             TimeUnit units,
                                             Object info)
   {
      long alarmTime = CurrentSimTime () + units.ConvertToUS (time);

      TimeoutID id = new TimeoutID ();
      id.time = alarmTime;
      id.info = info;
      id.client = client;
      id.internalWakeup = false;
      id.thread = null;

      AddTimeout (id);

      notifyAll ();     // wake up the timeoutThread

      return id;
   }

   /************************************************************
     Internal function to manipulate the event queue.  Try to be
     careful to make this a stable insertion.
    ************************************************************/
   private void AddTimeout (TimeoutID id)
   {
      int i = 0;
      while (i < pendingTimeouts.size () &&
             ((TimeoutID) pendingTimeouts.elementAt (i)).time <= id.time)
         ++i;

      pendingTimeouts.insertElementAt (id, i);
   }

   /************************************************************
     Given an ID from a previous SetTimeout,
             remove it from the queue if it is still pending.	  
    ************************************************************/
   public synchronized void CancelTimeout (TimeoutID id)
   {
      if (id == null)
         throw new NullPointerException ("Tried to cancel null timeout");
      
      if (pendingTimeouts == null)
         {
         System.out.println ("BAD: pendingTimeouts went null during execution");
         System.exit (1);
         }

      pendingTimeouts.removeElement (id);
   }



   /************************************************************
      Block the current thread on behalf of the client
      for a specified number
      of simulated time units.  The current thread must be registered,
      and must be currently blocking time.  
      The current thread must
      also hold the client's monitor, which will be released for
      the duration of the wait.
	
      @param client the client to wait on behalf of             
      @param time   time units to sleep
      @param units  units for time
      @throws IllegalTimerStateException if the current thread
              is not registered and blocking time
      @throws IllegalMonitorStateException if the current thread
              does not hold the client's monitor
    ************************************************************/
   public void WaitFor (TimerClient client, long time, TimeUnit units) 
                                 throws IllegalTimerStateException,
                                        IllegalMonitorStateException
   {
      long wakeupTime = CurrentSimTime () + units.ConvertToUS (time);
      WaitUntil (client, wakeupTime, US);
   }

   /************************************************************
      Block the current thread until a specified simulates time.
      The current thread must be registered,
      and must be currently blocking time.
      The current thread must
      also hold the client's monitor, which will be released for
      the duration of the wait.
	
      @param client the client to wait on behalf of             
      @param time   time units to sleep
      @param units  units for time
      @throws IllegalTimerStateException if the current thread
              is not registered and blocking time
      @throws IllegalMonitorStateException if the current thread
              does not hold the client's monitor
    ************************************************************/
   public void WaitUntil (TimerClient client, long time, TimeUnit units)
                                    throws IllegalTimerStateException,
                                           IllegalMonitorStateException
   {
      Thread current = Thread.currentThread ();
      if (!SimTimeIsBlockedFor (current))
         throw new IllegalTimerStateException ();

      long wakeupTime = units.ConvertToUS (time);

      // make sure we sleep for a little while
      if (wakeupTime <= CurrentSimTime ())
         wakeupTime = CurrentSimTime () + 1; 

      Debug.println ('t', "In WaitUntil on client " + client + 
                          ", about to grab timer lock");
      synchronized (this)
         {
         Debug.println ('t', "Got it.");
         TimeoutID id = new TimeoutID ();
         id.time = wakeupTime;
         id.info = null;
         id.client = client;
         id.internalWakeup = true;
         id.thread = current;

         AddTimeout (id);
         notifyAll ();    // wake up the run thread
         }

      while (CurrentSimTime () < wakeupTime)
         {
         Debug.println ('t', "In WaitUntil on client " + client + 
                             ", about to call Wait");
         client.Wait ();
         }
   }

   /************************************************************
      Tell the simulator that this thread cares about simulated
       time.  This allows the thread to use the function
      <tt>BlockSimTimeFor</tt>, <tt>UnblockSimTimeFor</tt> and
      <tt>SimTimeIsBlockedFor</tt>, as well as the monitor
      functions in <tt>TimerClient</tt>.
      <p>
      Simulated time advances only when all registered threads
      are not blocking sim time.  When first registered,
      a thread <i>is</i> blocking sim time,

      @param t the thread to register
      @throws IllegalTimerStateException if this thread is
            already registered.
    ************************************************************/
   public synchronized void RegisterThread (Thread t)
                                       throws IllegalTimerStateException
   {
      if (activeThreads.containsKey (t))
         throw new IllegalTimerStateException ();
      else
         {
         activeThreads.put (t, Boolean.FALSE);   // not waiting
         notifyAll ();
         }
   }

   /************************************************************
     Tell the simulator that this thread no
                     longer cares about simulated time.

      @param t the thread to register
      @throws IllegalTimerStateException if this thread is not
         currently registered
    ************************************************************/
   public synchronized void UnregisterThread (Thread t) 
                                       throws IllegalTimerStateException
   {
      if (!activeThreads.containsKey (t))
         throw new IllegalTimerStateException ();

      activeThreads.remove (t);

      notifyAll ();
      Debug.println ('t', "Just unregistered thread '" + t + "'");
   }

   /************************************************************
      Force simulated time to stall until this thread unblocks
      it or unregisters.  Threads should avoid sleeping while
      blocking sim time.  
      <p>
      In general it will probably not be
      necessary for any client to call this function, but it
      is available just in case.

      @param t the thread to block on behalf of
      @throws IllegalTimerStateException if this thread is not
         currently registered or is already blocking sim time
    ************************************************************/
   public synchronized void BlockSimTimeFor (Thread t)
                                        throws IllegalTimerStateException
   {
      if (!activeThreads.containsKey (t))
         throw new IllegalTimerStateException ();

      if (activeThreads.get (t) == Boolean.FALSE)
         throw new IllegalTimerStateException ();

      activeThreads.put (t, Boolean.FALSE);      

      notifyAll ();
   }

   /************************************************************
      Remove the blocking constraint on this thread.
      <p>
      In general it will probably not be
      necessary for any client to call this function, but it
      is available just in case.

      @param t the thread to unblock on behalf of
      @throws IllegalTimerStateException if this thread is not
         currently registered or is not already blocking sim time
    ************************************************************/
   public synchronized void UnblockSimTimeFor (Thread t)
                                       throws IllegalTimerStateException
   {
      if (!activeThreads.containsKey (t))
         throw new IllegalTimerStateException ();

      if (activeThreads.get (t) == Boolean.TRUE)
         throw new IllegalTimerStateException ();

      activeThreads.put (t, Boolean.TRUE);      

      notifyAll ();
   }


   /************************************************************
      Check whether this thread is blocking sim time.
      <p>
      In general it will probably not be
      necessary for any client to call this function, but it
      is available just in case.

      @param t the thread to check
      @throws IllegalTimerStateException if this thread is not
         currently registered.
      @return <tt>true</tt> if the thread is blocking sim time.
    ************************************************************/
   public synchronized boolean SimTimeIsBlockedFor (Thread t)
                                        throws IllegalTimerStateException
   {
      if (!activeThreads.containsKey (t))
         throw new IllegalTimerStateException ();

      return (activeThreads.get (t) == Boolean.FALSE);
   }


   /************************************************************
      main event processing loop: if we can advance time, then 
      advance to the next event and process it.  
      <p>Clients should not call this function.
    ************************************************************/
   public synchronized void run ()
   {
      Thread.currentThread ().setName ("Timeout processor");
      for (;;)
         {
         boolean timeAdvanced = false;

         boolean canGo = true;
         Enumeration e = activeThreads.elements ();
         while (e.hasMoreElements ())
            if (e.nextElement () == Boolean.FALSE)
               canGo = false;

         Debug.println ('t', "Top of process loop: canGo is " + canGo);
         Debug.print ('t', "Queue: ");
         Enumeration pendingEvents = pendingTimeouts.elements ();
         while (pendingEvents.hasMoreElements ())
            {
            TimeoutID data = (TimeoutID) pendingEvents.nextElement ();
            Debug.print ('t', "[" + data.time + " (");
            if (data.internalWakeup)
               Debug.print ('t', "w '" + data.thread + ")] ");
            else
               Debug.print ('t', "t '" + data.client + ")] ");
            }
         Debug.println ('t', "");

         while (pendingTimeouts.size () > 0 && canGo)
            {
            // All threads blocked, trigger next event
            TimeoutID id = (TimeoutID) pendingTimeouts.elementAt (0);
            pendingTimeouts.removeElementAt (0);
            if (simTime < id.time)
               {
               simTime = id.time;
               timeAdvanced = true;
               }

            Thread t = new TimeoutCallbackThread (this, id);
            RegisterThread (t);
            t.start ();

            canGo = true;
            e = activeThreads.elements ();
            while (e.hasMoreElements ())
               if (e.nextElement () == Boolean.FALSE)
                  canGo = false;

            Debug.print ('t', "canGo is " + canGo + "; Queue: ");
            pendingEvents = pendingTimeouts.elements ();
            while (pendingEvents.hasMoreElements ())
               {
               TimeoutID data = (TimeoutID) pendingEvents.nextElement ();
               Debug.print ('t', "[" + data.time + " (");
               if (data.internalWakeup)
                  Debug.print ('t', "w '" + data.thread + ")] ");
               else
                  Debug.print ('t', "t '" + data.client + ")] ");
               }
            Debug.println ('t', "");
            }

         if (timeAdvanced)
            notifyAll ();

         try
            {
            waiting = true;
            Debug.println ('t', "Timeout proceesor waiting.");
            wait ();
            waiting = false;
            }
         catch (InterruptedException ex)
            {
            }
         }
   }

   private long simTime;
   private Vector pendingTimeouts;
   private Hashtable activeThreads;
   private Thread timeoutThread;

   private boolean waiting;
   private boolean inTimeout;

   /************************************************************
      Print out the status of the timer, including event queue
      and registered threads.  Should be synchronized, but this
      is useful as a debugging tool when things deadlock, so
      it is left unsychronized.
    ************************************************************/
   public /*synchronized*/ void PrintStatus ()
   {
      try
         {
         System.out.print ("\nTimer stalled at time " + CurrentSimTime ()
                           + ": Current Timer Status:");
         if (waiting)
            System.out.println (" waiting");
         if (inTimeout)
            System.out.println (" in timeout");

         System.out.print ("Queue: ");
         Enumeration pendingEvents = pendingTimeouts.elements ();
         while (pendingEvents.hasMoreElements ())
            {
            TimeoutID data = (TimeoutID) pendingEvents.nextElement ();
            System.out.print ("[" + data.time + " (");
            if (data.internalWakeup)
               System.out.print ("w '" + data.thread + ")] ");
            else
               System.out.print ("t '" + data.client + ")] ");
            }
         System.out.println ("");
         System.out.println ("Active Threads: ");

         Enumeration e = activeThreads.keys ();
         while (e.hasMoreElements ())
            {
            Thread t = (Thread) e.nextElement ();
            System.out.print ("  " + t + ": ");
            if (activeThreads.get (t) == Boolean.TRUE)
               System.out.println (" not blocking sim time");
            else
               System.out.println (" blocking sim time");
            }

         Thread[] threads = new Thread[50];
         int count = Thread.currentThread ().enumerate (threads);
         System.out.println ("All threads:");
         for (int i=0; i<count; ++i)
            {
            System.out.println ("   '" + threads[i] + "'");
            }
         }
      catch (Exception e)
         {
         }
   }
}




//-----------------------------------------------------------------
// TimeoutCallbackThread: a thread to run timeout processing 
//                        in the background.  This allows
//                        timeouts to run without holding the
//                        Timer monitor, which is desireable for
//                        deadlock avoidance.
//-----------------------------------------------------------------
class TimeoutCallbackThread extends Thread
{
   // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
   // Set up the thread
   // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
   public TimeoutCallbackThread (Timer timer, TimeoutID id)
   {
      this.timer = timer;
      this.id = id;
   }

   private Timer timer;
   private TimeoutID id;

   // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
   // Run the timeout.  Note that we run registered to prevent 
   // time from advancing.
   // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
   public void run ()
   {
      Thread.currentThread ().setName ("Timeout run thread for client " + id.client);
      /* Already registered when created
         timer.RegisterThread (Thread.currentThread ());
      */

      if (id.internalWakeup)
         {
         synchronized (Timer.global)
            {
            id.client.NotifyThread (id.thread);
            }
         }
      else
         {
         synchronized (Timer.global)
            {
            id.client.Timeout (id.info);
            }
         }

      timer.UnregisterThread (Thread.currentThread ());
   }
}

//-----------------------------------------------------------------
// TimerMonitor: another thread which watches to see if the timer
//               appears stalled (i.e. deadlocked), and prints
//               out some status if it happens
//-----------------------------------------------------------------
class TimerMonitor extends Thread
{
   // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
   // Set up the timer to monitor
   // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
   public TimerMonitor (Timer timer)
   {
      this.timer = timer;
   }

   private Timer timer;

   // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
   // Wait for 2 real seconds, and if simulated time doesn't advance
   // assume deadlock.
   // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
   public void run ()
   {
      Thread.currentThread ().setName ("Timer monitor");

      long lastPrint = -1;

      for (;;)
         {
         long t1 = timer.CurrentSimTime ();

         try
            {
            sleep (2000);
            }
         catch (InterruptedException e)
            {
            }

         long t2 = timer.CurrentSimTime ();
         if (t1 == t2 && t2 != lastPrint)
            {
            lastPrint = t2;
            timer.PrintStatus ();
            }
         }
   }
}

