//-----------------------------------------------------------------
// CSE 461 Project 1:                        Timer implementation
//
// implementation of the simulated time clock
// and other time-related functions.
//-----------------------------------------------------------------
package timer;

import java.io.*;
import java.util.*;
import debug.*;

/***************************************************************** 
 *****************************************************************/
   /************************************************************
    ************************************************************/


/***************************************************************** 
   A class to describe clients of the timer.  All
   classes wishing to use the features of the timer
   will need to <tt>extend</tt> this class, and most 
   will need to override the <tt>Timeout</tt> function 
   as well.  The purpose of this class is twofold:
<ol>
<li> To describe the callback interface for timeouts.  The method
     <tt>Timeout</tt> will be called (with the data object associated
     with the timeout) when timeouts fire.  Clients using timeouts will
     therefore need to override this function.
<li> To implement synchronization correctly with respect to the
     timer.  This class implements capitalized versions of the 
     standard Java monitor functions.  Clients of the timer will
     almost certainly want to use those functions in place of the 
     builtin lowercase versions.  In addition to allowing time to
     advance while a thread is blocked, they implement true 
     FIFO queuing of <tt>Wait()</tt>ing threads, so that 
     <tt>Notify</tt> wakes the oldest waiting thread.
     <p>
     Also note: these monitor functions all operate on the
     global timer's monitor.
</ol>
 *****************************************************************/

public class TimerClient
{
   /************************************************************
      Set up the monitor state for the object
    ************************************************************/
   public TimerClient ()
   {
      waitQueue = new Vector ();
   }

   /************************************************************
      the callback for timeouts.  When a timeout fires,
      this method on the client will be called.  The default
      implementation does nothing.

      @param info the object which was passed to <tt>SetTimeout</tt>
         when this timeout was set
      @see Timer#SetTimeout
    ************************************************************/
   public void Timeout (Object info)
   {
      // This space intentionally left bank
   }

   /************************************************************
      Block the currently executing thread indefiniately until
      <tt>Notify</tt> or <tt>NotifyAll</tt> is called.  The 
      current thread must hold the timer's monitor and be 
      registered as blocking simulated time.  While the thread
      is blocked both the monitor and simulated time will be 
      released.

      @throws IllegalMonitorStateException if the current
              thread does not hold the timer's monitor
      @throws IllegalTimerStateException if the current
              thread is not blocking simulated time.
    ************************************************************/
   public void Wait () throws IllegalMonitorStateException,
                              IllegalTimerStateException
   {
      Thread waiter;
      synchronized (Timer.global)
         {
         waiter = Thread.currentThread ();
         waitQueue.addElement (waiter);
         Timer.global.UnblockSimTimeFor (waiter);
         }

      while (!Timer.global.SimTimeIsBlockedFor (waiter))
         try
            {
            Timer.global.wait ();
            }
         catch (InterruptedException e)
            {
            }
   }

   /************************************************************
      <b>This function not yet implemented</b>
      <p>
      Block the currently executing thread for at most
      <tt>us</tt> microseconds or until
      <tt>Notify</tt> or <tt>NotifyAll</tt> is called.  The 
      current thread must hold the timer's monitor and be 
      registered as blocking simulated time.  While the thread
      is blocked both the monitor and simulated time will be 
      released.

      @param us the maximum wait time in simulated microseconds

      @throws IllegalMonitorStateException if the current
              thread does not hold the timer's monitor
      @throws IllegalTimerStateException if the current
              thread is not blocking simulated time.
    ************************************************************/
   public void Wait (long us) throws IllegalMonitorStateException,
                                     IllegalTimerStateException
   {
      throw new IllegalTimerStateException ();
   }

   /************************************************************
      Wake up the first waiting thread on the queue, if any.
      Does not cause that thread to run immediately, just makes
      it runnable.

      @throws IllegalMonitorStateException if the current
              thread does not hold the timer's monitor
      @throws IllegalTimerStateException if the current
              thread is not blocking simulated time.
    ************************************************************/
   public void Notify () throws IllegalMonitorStateException,
                                IllegalTimerStateException
   {
      boolean go = false;

      synchronized (Timer.global)
         {
         if (waitQueue.size () > 0)
            {
            Thread waiter = (Thread) waitQueue.elementAt (0);
            waitQueue.removeElementAt (0);
            Timer.global.BlockSimTimeFor (waiter);
            go = true;
            }
         }

      if (go)
         Timer.global.notifyAll ();
   }

   /************************************************************
      A new monitor twist:
      Wake up a specific waiting thread on the queue, if present.
      Does not cause that thread to run immediately, just makes
      it runnable.

      @param t the thread to wake

      @throws IllegalMonitorStateException if the current
              thread does not hold the timer's monitor
      @throws IllegalTimerStateException if the current
              thread is not blocking simulated time.
    ************************************************************/
   public void NotifyThread (Thread t) throws IllegalMonitorStateException,
                                              IllegalTimerStateException
   {
      boolean go = false;

      synchronized (Timer.global)
         {
         for (int i = 0; i < waitQueue.size (); ++i)
         if (waitQueue.size () > 0)
            {
            Thread waiter = (Thread) waitQueue.elementAt (i);
            if (waiter == t)
               {
               waitQueue.removeElementAt (i);
               Timer.global.BlockSimTimeFor (waiter);
               go = true;
               break;
               }
            }
         }

      if (go)
         Timer.global.notifyAll ();
   }

   /************************************************************
      Wake up all waiting threads on the queue, if any.
      Does not cause those thread sto run immediately, just makes
      them runnable.

      @throws IllegalMonitorStateException if the current
              thread does not hold the timer's monitor
      @throws IllegalTimerStateException if the current
              thread is not blocking simulated time.
    ************************************************************/
   public void NotifyAll () throws IllegalMonitorStateException,
                                   IllegalTimerStateException
   {
      synchronized (Timer.global)
         {
         while (waitQueue.size () > 0)
            {
            Thread waiter = (Thread) waitQueue.elementAt (0);
            waitQueue.removeElementAt (0);
            Timer.global.BlockSimTimeFor (waiter);
            }
         }

      Timer.global.notifyAll ();
   }

   /************************************************************
      The queue of waiting threads.
    ************************************************************/
   private Vector waitQueue;
}
